/**************************************************************************

	ns

	main.c

	Copyright (C) 1996-2002 by Tomoaki Nakashima. All rights reserved.
		http://www.nakka.com/
		nakka@nakka.com

**************************************************************************/

/**************************************************************************
	Include Files
**************************************************************************/

#include <windows.h>
#include <stdio.h>

#include "script.h"
#include "script_regex.h"
#include "script_string.h"
#include "script_memory.h"
#include "script_utility.h"


/**************************************************************************
	Define
**************************************************************************/

#define BUF_SIZE	256
#define LINE_SIZE	32768


/**************************************************************************
	Global Variables
**************************************************************************/

TCHAR AppDir[BUF_SIZE];


/******************************************************************************

	DebugNFA

	NFA(K\)̃fobO\

******************************************************************************/

static void DebugNFA(rege_nfa_t *nfa)
{
	int i;
	nlist_t *p;
	char *s, *r, *val;

	for(i = 0; i < nfa->nstate; i++){
		printf("  state %3d: ", i);
		if(nfa->states[i] == NULL){
			if(nfa->exit == i){
				printf("ACCEPT\n");
			}else{
				printf("ERROR\n");
			}
			continue;
		}
		for(p = nfa->states[i]; p != NULL; p = p->next){
			if(p->op == 3){
				printf("([%s] . %d) ", p->u.chr_class, p->to);
#ifdef REGVARIABLE
			}else if(p->op == 11){
				if(p->u.v.st_var != NULL){
					s = p->u.v.st_var;
				}else{
					s = p->u.v.end_var;
				}
				for(r = s + 1; *r != '\0' && *r != '%'; r++);
				if(*r == '%') r++;
				val = alloc_copy_n(s, r - s + 1);
				printf("(%s . %d) ", val, p->to);
				mem_free(&val);
#endif
			}else{
				printf("(");
				if(p->eps == -1){
					switch(p->op)
					{
					case 1:
						printf("\"HEAD\"");
						break;

					case 2:
						printf("\"TAIL\"");
						break;

					case 4:
						printf("\"CONCAT\"");
						break;

					case 5:
						printf("\"UNION\"");
						break;

					case 6:
					case 7:
						printf("\"CLOSURE\"");
						break;

					case 8:
						printf("\"EMPTY\"");
						break;

					case 9:
						printf("\"OPEN %d\"", p->u.i);
						break;

					case 10:
						printf("\"CLOSE %d\"", p->u.i);
						break;

					default:
						printf("?");
						break;
					}
				}else{
					switch(p->u.c.c1)
					{
					case '\r':
						printf("\\r");
						break;

					case '\n':
						printf("\\n");
						break;

					case '\t':
						printf("\\t");
						break;

					default:
						printf("%c", p->u.c.c1);
						if(p->u.c.c2 != 0){
							printf("%c", p->u.c.c2);
						}
					}
				}
				printf(" . %d) ", p->to);
			}
		}
		printf("\n");
	}
}


/******************************************************************************

	DebugToken

	͖؂̃fobO\

******************************************************************************/

static void DebugToken(TOKEN *tk, int lv, BOOL shownfa)
{
	char *p;
	int i;

	if(tk == NULL) return;

	for(i = 0; i < lv; i++){
		printf("    ");
	}
	if(tk->Type == SYM_CONST){
		printf("%ld", tk->i);
	}else if(tk->buf != NULL){
		if(tk->Type == SYM_STRING){
			printf("\"");
			for(p = tk->buf; *p != '\0'; p++){
				switch(*p)
				{
				case '\r':
					printf("\\r");
					break;

				case '\n':
					printf("\\n");
					break;

				default:
					printf("%c", *p);
					break;
				}
			}
			printf("\"");
		}else{
			printf("%s", tk->buf);
		}
	}
	switch(tk->Type)
	{
	case SYM_FUNC:
		printf(" (function)");
		break;
	}
	printf("\n");

	if(shownfa == TRUE && tk->nfa != NULL){
		printf("-- NFA\n");
		DebugNFA(tk->nfa);
		printf("--\n");
	}

	DebugToken(tk->target, lv + 1, shownfa);
	DebugToken(tk->next, lv, shownfa);
}


/******************************************************************************

	DebugScriptInfo

	XNvg̃fobO\

******************************************************************************/

static void DebugScriptInfo(SCRIPTINFO *sci, BOOL shownfa)
{
	while(sci != NULL){
		if(sci->name != NULL){
			printf("Script: %s\n", sci->name);
		}
		DebugToken(sci->tk, 1, shownfa);
		sci = sci->next;
	}
}


/******************************************************************************

	LineExecLoop

	Ps͂́Ass

******************************************************************************/

static void LineExecLoop(BOOL debug, BOOL shownfa)
{
	SCRIPTINFO *sci;
	EXECINFO ei;
	VALUEINFO *vi;
	VALUEINFO *svi = NULL;
	TOKEN *tk;
	char buf[LINE_SIZE];
	int ret = 0;

	sci = mem_alloc(sizeof(SCRIPTINFO));
	InitializeScriptInfo(sci, FALSE);

	ZeroMemory(&ei, sizeof(EXECINFO));
	ei.Name = "stdin";
	ei.sci = sci;

	while(fgets(buf, LINE_SIZE - 1, stdin) != NULL){
		//
		sci->buf = buf;
		tk = ParseSentence(&ei, buf, 1);
		if(tk == NULL){
			continue;
		}
		if(debug == TRUE){
			//͖ؕ\
			DebugToken(tk, 1, shownfa);
		}else{
			//s
			svi = NULL;
			ret = ExecSentense(&ei, tk, NULL, &svi, NULL);
		}
		FreeToken(tk);
		if(ret == -3){
			break;
		}

		vi = svi;
		while(vi != NULL){
			if(vi->v != NULL){
				if(vi->v->Type == TYPE_INTEGER){
					printf("%ld\n", vi->v->u.iValue);
				}else{
					printf("%s\n", vi->v->u.Value);
				}
			}
			vi = vi->next;
		}

		FreeValueList(svi);

	}
	FreeExecInfo(&ei);
	sci->buf = NULL;
	FreeScriptInfo(sci);
}


/******************************************************************************

	main

	C

******************************************************************************/

int __cdecl main(int argc, char **argv)
{
	SCRIPTINFO *ScriptInfo;
	EXECINFO ei;
	VALUEINFO *vi, *pvi;
	TCHAR fname[BUF_SIZE];
	TCHAR *p, *r;
	int i = 1;
	char *c;
	BOOL op_debug = FALSE;
	BOOL op_nfa = FALSE;
	BOOL op_explicit = FALSE;

	if(argc > i && (*(argv[i]) == '/' || *(argv[i]) == '-')){
		//help
		for(c = argv[i]; *c != '\0' && *c != '?'; c++);
		if(*c != '\0'){
#ifdef LC_JP
			printf("ns [/edn] [file.ns] [arg1[ arg2...]]\n");
			printf("\n");
			printf("  e\t\tϐ錾 (ʏs)\n");
			printf("  d\t\t͖؂\\n");
			printf("  n\t\tNFA(K\)\\n");
			printf("\n");
			printf("  d ̎w肪ꍇ͎s[hɂȂ\n");
			printf("  n  d w肵Ƃ̂ݗL\n");
			printf("\n");
			printf("  file.ns\tsA͂sXNvgt@C\n");
			printf("         \tt@C̎w肪ꍇ̓Css\n");
			printf("\n");
			printf("  arg1\t\tXNvgɓn\n");
			printf("\n");
#else
			printf("ns [/edn] [file.ns] [arg1[ arg2...]]\n");
			printf("\n");
			printf("  e\t\tExplicit\n");
			printf("  d\t\tShow token (parse tree)\n");
			printf("  n\t\tShow NFA (regular expression)\n");
			printf("\n");
			printf("  file.ns\tExecution script file\n");
			printf("\n");
			printf("  arg1\t\tCommand line\n");
			printf("\n");
#endif
			return 0;
		}
		//debug
		for(c = argv[i]; *c != '\0' && *c != 'd' && *c != 'D'; c++);
		if(*c != '\0'){
			op_debug = TRUE;
		}
		//show NFA
		for(c = argv[i]; *c != '\0' && *c != 'n' && *c != 'N'; c++);
		if(*c != '\0'){
			op_nfa = TRUE;
		}
		//explicit
		for(c = argv[i]; *c != '\0' && *c != 'e' && *c != 'E'; c++);
		if(*c != '\0'){
			op_explicit = TRUE;
		}
		i++;
	}

	if(argc <= i){
		//1ss[h
		InitializeScript();
		LineExecLoop(op_debug, op_nfa);
		TerminateScript();
		EndScript();
#ifdef _DEBUG
	mem_debug();
#endif
		return 0;
	}


	p = char2tchar(argv[i]);
	lstrcpy(AppDir, p);
	mem_free(&p);
	i++;

	//t@C̃fBNg擾
	for(p = r = AppDir; *p != TEXT('\0'); p++){
#ifdef UNICODE
		if(*p == TEXT('\\') || *p == TEXT('/')){
			r = p;
		}
#else
		if(IsDBCSLeadByte(*p) == TRUE && *(p + 1) != '\0'){
			p++;
		}else if(*p == '\\' || *p == '/'){
			r = p;
		}
#endif
	}
	//t@C擾
	if(r != AppDir){
		*r = TEXT('\0');
		lstrcpy(fname, r + 1);
		lstrcat(AppDir, TEXT("\\"));
	}else{
		lstrcpy(fname, AppDir);
		GetCurrentDirectory(BUF_SIZE + 1, AppDir);
		lstrcat(AppDir, TEXT("\\"));
	}
	SetCurrentDirectory(AppDir);

	//
	InitializeScript();

	//ǂݍ
	ScriptInfo = mem_alloc(sizeof(SCRIPTINFO));
	if(ScriptInfo == NULL){
		return -1;
	}
	InitializeScriptInfo(ScriptInfo, op_explicit);

#ifdef UNICODE
	{
		char *c1, *c2;
		c1 = tchar2char(AppDir);
		c2 = tchar2char(fname);
		ReadScriptFile(ScriptInfo, c1, c2);
		mem_free(&c1);
		mem_free(&c2);
	}
#else
	ReadScriptFile(ScriptInfo, AppDir, fname);
#endif
	ExecGlobalArea(ScriptInfo);

	if(op_debug == TRUE){
		//͖ؕ\
		DebugScriptInfo(ScriptInfo, op_nfa);
	}else{
		//
		if(argc <= i){
			pvi = NULL;
		}else{
			vi = pvi = StringToVariable(NULL, argv[i]);
			for(i++; i < argc; i++){
				vi = vi->next = StringToVariable(NULL, argv[i]);
			}
		}

		//s
		ZeroMemory(&ei, sizeof(EXECINFO));
		ei.sci = ScriptInfo->next;
		vi = ExecFunction(&ei, NULL, "main", pvi, FALSE);
		if(vi != NULL && (int)vi != -1){
			FreeValue(vi);
		}
		FreeValueList(pvi);
	}
	TerminateScript();
	FreeScriptInfo(ScriptInfo);
	EndScript();
#ifdef _DEBUG
	mem_debug();
#endif
	return 0;
}
/* End of source */
